/* message.c      message passing routines

   copyright(c)  Prabhas Chongstitvatana     	20 Jan 97

   add xTmSend, xTmReceive			21 Aug 98

 */

#include "interp.h"

#define   PRIVATE        static

#define   mailmbox(p)              pcb[p].mbox
#define   mailawait(p)             pcb[p].await
#define   mailid(m)                mm[m].id
#define   mailmessage(m)           mm[m].message
#define   mailnext(m)              mm[m].next
#define   setmail(m,pid,msg)       mm[m].id = (pid); mm[m].message = (msg)

PRIVATE mdata  mm[maxmail];
PRIVATE mailp  freemail;

PRIVATE mailp newmail()
{
  mailp m;
  if( freemail == nil ) error("ran out of mail space");
  m = freemail;
  freemail = mailnext(m);
  return m;
}

PRIVATE disposemail(m)
mailp m;
{
  mailnext(m) = freemail;
  freemail = m;
}

/* retreive a mail of id from the current p mbox, await */
PRIVATE mailp getmail(mp,id)
mailp *mp;
pdesc id;
{
  mailp a,b;
  a = *mp;
  b = a;
  while ( a != nil ) {
    if( mailid(a) == id ) {
      if( a == b )    /* found at head of the list */
	*mp = mailnext(a);
      else
	mailnext(b) = mailnext(a);
      mailnext(a) = nil;
      return a;
    }
    b = a;
    a = mailnext(a);
  }
  return nil;
}


/* push the message to rec's variable */
PRIVATE pushmailp(pdesc p, int a)
{
  M[ M[pcb[p].sp + 2] ] = a;
}

PRIVATE void sendmail( mailp *mp, int mess )
{
  mailp m;
  m = newmail();
  setmail(m,activep,mess);
  mailnext(m) = *mp;		/* insert mail */
  *mp = m;
}

initmail()
{
  int i;
  for( i = 1; i < maxmail-1; i++ )
    mm[i].next = i+1;
  freemail = 1;
}

void undombox( pdesc p2, pdesc p )
{
  disposemail( getmail( &(mailmbox(p2)), p));
}

void undoawait( pdesc p2, pdesc p )
{
  disposemail( getmail( &(mailawait(p2)), p));
}

xSend()
{
  mailp m;
  pdesc p;
  int mess;
  mess = TOS; p = NEXT; pop2; ip++;
  if( (m = getmail( &(mailawait(activep)), p)) != nil ) {
    pushmailp(p,mess);
    disposemail(m);
    awakep(p);
  } else {
    sendmail( &(mailmbox(p)),mess);
    blockp(SEND);
  }
}

xReceive()
{
  mailp m;
  pdesc p;
  int i;
  i = pop(); p = pop(); ip++;
  if( (m = getmail( &(mailmbox(activep)), p)) != nil ) {
    M[i] = mailmessage(m);
    disposemail(m);
    awakep(p);
  } else {
    sendmail( &(mailawait(p)), 0);
    blockp(RECEIVE);
  }
}

xTmSend()
{
  mailp m;
  pdesc p;
  int mess;
  int t;
  t = pop(); mess = TOS; p = NEXT; pop2; ip++;
  if( (m = getmail( &(mailawait(activep)), p)) != nil ) {
    pushmailp(p,mess);
    disposemail(m);
    awakep(p);
  } else {
    sendmail( &(mailmbox(p)), mess);
    blockp(TMSEND);
    starttimer(t,p);
  }
}

xTmReceive()
{
  mailp m;
  pdesc p;
  int ads, t;
  t = pop(); ads = TOS; p = NEXT; pop2; ip++;
  if( (m = getmail( &(mailmbox(activep)), p)) != nil ) {
    M[ads] = mailmessage(m);
    disposemail(m);
    awakep(p);
  } else {
    sendmail( &(mailawait(p)), 0);
    blockp(TMRECEIVE);
    starttimer(t,p);
  }
}

